/*
* Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <linux/version.h>
#include <linux/uaccess.h>
#include <linux/kprobes.h>
#include <linux/ftrace.h>
#include <linux/kallsyms.h>
#include <asm/ptrace.h>
#if !defined(CONFIG_LOONGARCH)
#include <asm/insn.h>
#endif
#include "inno_insn.h"
#include "inno_misc.h"

#if defined(CONFIG_ARM64)

#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0))
static const int aarch64_insn_encoding_class2[] = {
	AARCH64_INSN_CLS_UNKNOWN,
	AARCH64_INSN_CLS_UNKNOWN,
	AARCH64_INSN_CLS_UNKNOWN,
	AARCH64_INSN_CLS_UNKNOWN,
	AARCH64_INSN_CLS_LDST,
	AARCH64_INSN_CLS_DP_REG,
	AARCH64_INSN_CLS_LDST,
	AARCH64_INSN_CLS_DP_FPSIMD,
	AARCH64_INSN_CLS_DP_IMM,
	AARCH64_INSN_CLS_DP_IMM,
	AARCH64_INSN_CLS_BR_SYS,
	AARCH64_INSN_CLS_BR_SYS,
	AARCH64_INSN_CLS_LDST,
	AARCH64_INSN_CLS_DP_REG,
	AARCH64_INSN_CLS_LDST,
	AARCH64_INSN_CLS_DP_FPSIMD,
};
#endif

int aarch64_get_insn_class2(u32 insn)
{
	int res = -1;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0))
	switch (aarch64_insn_encoding_class2[(insn >> 25) & 0xf]) {
		case AARCH64_INSN_CLS_BR_SYS:
			res = INNO_AARCH64_INSN_CLS_BR_SYS;
			break;
		case AARCH64_INSN_CLS_LDST:
			res = INNO_AARCH64_INSN_CLS_LDST;
			break;
		default:
				res = -1;
	}
#endif
	return res;
}
INNO_EXT_SYM(aarch64_get_insn_class2);

int inno_read_cpuid(void)
{
	return read_cpuid(DCZID_EL0);
}
INNO_EXT_SYM(inno_read_cpuid);
#endif

int inno_user_mode(inno_pt_regs *regs)
{
	return user_mode((struct pt_regs *)regs);
}
INNO_EXT_SYM(inno_user_mode);

void inno_set_pt_regs_sp(inno_pt_regs *regs, u64 sp)
{
#ifdef CONFIG_ARM64
	struct pt_regs *pt_regs = (struct pt_regs *)regs;
	pt_regs->sp = sp;
#endif
}
INNO_EXT_SYM(inno_set_pt_regs_sp);

void inno_instruction_pointer_set(inno_pt_regs *regs, unsigned long val)
{
	instruction_pointer_set((struct pt_regs *)regs, val);
}
INNO_EXT_SYM(inno_instruction_pointer_set);

unsigned long inno_instruction_pointer(inno_pt_regs *regs)
{
	return instruction_pointer((struct pt_regs *)regs);
}
INNO_EXT_SYM(inno_instruction_pointer);

u64 inno_regs_get_register(inno_pt_regs *regs, unsigned int offset)
{
	return regs_get_register((struct pt_regs *)regs, offset);
}
INNO_EXT_SYM(inno_regs_get_register);

void inno_pt_regs_write_reg(inno_pt_regs *regs, int r, unsigned long val)
{
#ifdef CONFIG_ARM64
	pt_regs_write_reg((struct pt_regs *)regs, r, val);
#endif
}
INNO_EXT_SYM(inno_pt_regs_write_reg);

unsigned long inno_pt_regs_read_reg(inno_pt_regs *regs, int r)
{
#ifdef CONFIG_ARM64
	return pt_regs_read_reg((struct pt_regs *)regs, r);
#else
	return 0;
#endif
}
INNO_EXT_SYM(inno_pt_regs_read_reg);

long inno_probe_kernel_read(void *dst, const void *src, size_t size)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
	return probe_kernel_read(dst, src, size);
#else
	return copy_from_kernel_nofault(dst, src, size);
#endif
}
INNO_EXT_SYM(inno_probe_kernel_read);

static struct inno_fixup_kprobe {
#ifdef CONFIG_KPROBES
	struct kprobe kp;
#elif defined(CONFIG_FTRACE)
	struct ftrace_ops ops;
#endif
	int (*func)(unsigned long addr, inno_pt_regs *regs);
} _inno_fixup_kprobe;

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,15,86)
static int _inno_fixup_alignment_fun(unsigned long addr, unsigned int esr, struct pt_regs *regs)
#else
static int _inno_fixup_alignment_fun(unsigned long addr, unsigned long esr, struct pt_regs *regs)
#endif
{
	return _inno_fixup_kprobe.func(addr, regs);
}

#ifdef CONFIG_KPROBES
static int inno_fixup_alignment_pre(struct kprobe *p, struct pt_regs *regs) {
#elif defined(CONFIG_FTRACE)
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,11,0)
static void notrace inno_ftrace_callback(unsigned long ip, unsigned long parent_ip,
                                struct ftrace_ops *ops, struct pt_regs *regs) {
#else
static void notrace inno_ftrace_callback(unsigned long ip, unsigned long parent_ip,
                                struct ftrace_ops *ops, struct ftrace_regs *fregs) {
	struct pt_regs *regs = ftrace_get_regs(fregs);
#endif
#endif
	instruction_pointer_set(regs, (unsigned long)_inno_fixup_alignment_fun);
#ifdef CONFIG_KPROBES
	return 1;
#endif
}

void fixup_alignment_kprobe_init(inno_kprobe **inno_kp, char *name, int (*func)(unsigned long addr, inno_pt_regs *regs))
{
	_inno_fixup_kprobe.func = func;
#ifdef CONFIG_KPROBES
	_inno_fixup_kprobe.kp.symbol_name = name;
	_inno_fixup_kprobe.kp.pre_handler = inno_fixup_alignment_pre;
	*inno_kp = &(_inno_fixup_kprobe.kp);
#elif defined(CONFIG_FTRACE)
	_inno_fixup_kprobe.ops.func = inno_ftrace_callback;
	_inno_fixup_kprobe.ops.flags = FTRACE_OPS_FL_SAVE_REGS| FTRACE_OPS_FL_RECURSION_SAFE | FTRACE_OPS_FL_IPMODIFY;
#endif
}
INNO_EXT_SYM(fixup_alignment_kprobe_init);

int inno_register_kprobe(inno_kprobe *kp, char *name)
{
	int ret = 0;
#ifdef CONFIG_KPROBES
	ret = register_kprobe((struct kprobe *)kp);
#elif defined(CONFIG_FTRACE)
    ret = ftrace_set_filter(&(_inno_fixup_kprobe.ops), name, strlen(name), 0);

    ret = register_ftrace_function(&(_inno_fixup_kprobe.ops));
#endif
	return ret;
}
INNO_EXT_SYM(inno_register_kprobe);

void inno_unregister_kprobe(inno_kprobe *kp)
{
#ifdef CONFIG_KPROBES
	unregister_kprobe((struct kprobe *)kp);
#elif defined(CONFIG_FTRACE)
    unregister_ftrace_function(&(_inno_fixup_kprobe.ops));
#endif
}
INNO_EXT_SYM(inno_unregister_kprobe);
